# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Robert Schenk, <robert.schenk@uni-leipzig.de>
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
VKFilm (Postdam), #8575, about 120000 records.

Config
------

[vkfilm]

file = /path/to/file
"""

import luigi

from gluish.utils import shellout
from siskin.task import DefaultTask


class VKFilmTask(DefaultTask):
    TAG = '127'


class VKFilmFile(VKFilmTask, luigi.ExternalTask):
    """
    Manually downloaded MAB dump, #8575#note-7.
    """

    def output(self):
        return luigi.LocalTarget(path=self.config.get('vkfilm', 'file'))


class VKFilmMarc(VKFilmTask):
    """
    Convert to some MARC XML via metafacture. Custom morphs and flux are
    kept in assets/127.
    """

    def requires(self):
        return VKFilmFile()

    def run(self):
        output = shellout("""flux.sh {flux} in={input} > {output}""",
                          flux=self.assets("127/127.flux"), input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='fincmarc.xml'))



# coding: utf-8
# pylint: disable=C0301,C0103,E1101

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Experimental tasks for generic OAI sources. Works fine, if data source is somewhat standard-conform.

On the command line:

    $ taskdo OAIIntermediateSchema --url http://example.com/oai --prefix oai_dc --sid 123 --collection '"Stellar Articles"' --format ElectronicArticle

Other OAI considerations:

Decouple harvesting, extend metha, so it can harvest endpoints on autopilot,
place files into a special folder and configure this folder in siskin. Then
just require the already prepared file:

    def requires(self):
        return IntermediateSchemaFromOAI(endpoint="...", format="...")


"""

import datetime

import luigi

from gluish.common import Executable
from gluish.utils import shellout
from siskin.task import DefaultTask


class OAIBaseTask(DefaultTask):
    """
    Base task for generic OAI tasks.
    """
    TAG = "oai"


class OAICombine(OAIBaseTask):
    """
    Combine harvested files for a data source into a single file.
    """
    date = luigi.DateParameter(default=datetime.date.today())
    url = luigi.Parameter(default="http://export.arxiv.org/oai2", significant=False)
    prefix = luigi.Parameter(default="oai_dc", significant=False)

    def requires(self):
        return [
            Executable(name='metha-sync', message='https://github.com/miku/metha'),
            Executable(name='pigz', message='http://zlib.net/pigz/'),
        ]

    def run(self):
        shellout("METHA_DIR={dir} metha-sync -format {prefix} {url}", prefix=self.prefix, url=self.url, dir=self.config.get('core', 'metha-dir'))
        output = shellout("METHA_DIR={dir} metha-cat -root Records -format {prefix} {url} | pigz -c > {output}",
                          prefix=self.prefix, url=self.url, dir=self.config.get('core', 'metha-dir'))
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="xml.gz"))


class OAIIntermediateSchema(OAIBaseTask):
    """
    Convert OAI via flux, pass SID and collection as arguments.
    """
    date = luigi.DateParameter(default=datetime.date.today())
    url = luigi.Parameter(significant=False)
    prefix = luigi.Parameter(significant=False)

    sid = luigi.Parameter(description="internal data source id")
    collection = luigi.Parameter(description="internal collection name")
    format = luigi.Parameter(default='ElectronicArticle')

    def requires(self):
        return OAICombine(date=self.date, url=self.url, prefix=self.prefix)

    def run(self):
        mapdir = 'file:///%s' % self.assets("maps/")
        output = shellout("""flux.sh {flux}
                             MAP_DIR={mapdir}
                             in={input}
                             sid={sid}
                             mega_collection={collection}
                             format={format} | pigz -c > {output}""",
                          flux=self.assets("oai/flux.flux"), mapdir=mapdir, sid=self.sid,
                          collection=self.collection, format=self.format, input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz', digest=True))



#!/usr/bin/env python3
# coding: utf-8
# pylint: disable=C0103

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Robert Schenk, <robert.schenk@uni-leipzig.de>
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar. If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>


# Contains a collection of utilities for parsing and analysing xml files
# Current functions: fieldlist, print as json, validate xml
# To do: sort tags, count records, count doublet


import collections
import argparse
import xml.sax
import json
import os
import re
import subprocess


parser = argparse.ArgumentParser()
parser.add_argument("-v",
                    action="version",
                    help="show version",
                    version="0.0.1")
parser.add_argument("-f",
                    dest="inputfile",
                    help="file to parse",
                    metavar="filename")
parser.add_argument("-t",
                    dest="task",
                    help="task to do (list_fields, count_records)",
                    metavar="task")
args = parser.parse_args()

if args.task == "list_fields":

    class RecordHandler(xml.sax.ContentHandler):

        def __init__(self):
            self.fieldlist = {}
            self.current = (None, None)

        def startElement(self, name, attr):
            self.current = (name, attr)

        def characters(self, value):
            name = self.current[0]
            value = value.strip()
            if value:
                if name not in self.fieldlist:
                    self.fieldlist[name] = value

    parser = xml.sax.make_parser()
    handler = RecordHandler()
    parser.setContentHandler(handler)
    parser.parse(open(args.inputfile, "r"))

    fieldlist = handler.fieldlist
    for key, value in fieldlist.items():
        print(key + " = " + value)

elif args.task == "print_json":

    class RecordHandler(xml.sax.ContentHandler):

        def __init__(self):
            self.data = collections.defaultdict(list)
            self.current = (None, None)

        def startElement(self, name, attr):
            self.current = (name, attr)

        def endElement(self, name):
            print(json.dumps(self.data))
            self.data.clear()

        def characters(self, value):
            name = self.current[0]
            value = value.strip()
            if value:
                self.data[name].append(value)

    parser = xml.sax.make_parser()
    handler = RecordHandler()
    parser.setContentHandler(handler)
    parser.parse(open(args.inputfile, "r"))

elif args.task == "count_records":

    file = open(args.inputfile, "r", encoding="utf-8")
    for i, line in enumerate(file, start=0):
        if i == 5:  # diverse Kopfzeilen und Dokumententypdefinitionen werden übersprungen
            regexp = re.match("<(.*?)>.*</(.*)>", line)
            if regexp:
                starttag, endtag = regexp.groups()
                if starttag == endtag:
                    command = 'grep -c "</%s>" %s' % (endtag, args.inputfile)
                    os.system(command)
            else:
                pass

elif args.task == "validate_xml":
    os.system("xmllint --format %s" % args.inputfile)



# coding: utf-8
# pylint: disable=C0301,E1101,W0622,C0103
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                    The Finc Authors, http://finc.info
#                    Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

#
# Microsoft Academic Graph
#
# TOS: http://research.microsoft.com/en-us/um/redmond/projects/mag/license.txt
#
# Links:
#
# * http://research.microsoft.com/en-us/projects/mag/
# * https://academicgraph.blob.core.windows.net/graph/index.html?eulaaccept=on
#
# Number of papers as of November 2015: 120887833, ~37M with DOI
#
# Schema
# Affiliations
#
#     Affiliation ID
#     Affiliation name
#
# Authors
#
#     Author ID
#     Author name
#
# Conference
#
#     Conference ID
#     Short name (abbreviation)
#     Full name
#
# ConferenceInstances
#
#     Conference ID
#     Conference instance ID
#     Short name (abbreviation)
#     Full name
#     Location
#     Official conference URL
#     Conference start date
#     Conference end date
#     Conference abstract registration date
#     Conference submission deadline date
#     Conference notification due date
#     Conference final version due date
#
# FieldsOfStudy
#
#     Field of study ID
#     Field of study name
#
# Journals
#
#     Journal ID
#     Journal name
#
# Papers
#
#     Paper ID
#     Original paper title
#     Normalized paper title
#     Paper publish year
#     Paper publish date
#     Paper Document Object Identifier (DOI)
#     Original venue name
#     Normalized venue name
#     Journal ID mapped to venue name
#     Conference series ID mapped to venue name
#     Paper rank
#
# PaperAuthorAffiliations
#
#     Paper ID
#     Author ID
#     Affiliation ID
#     Original affiliation name
#     Normalized affiliation name
#     Author sequence number
#
# PaperKeywords
#
#     Paper ID
#     Keyword name
#     Field of study ID mapped to keyword
#
# PaperReferences
#
#     Paper ID
#     Paper reference ID
#
# PaperUrls
#
#     Paper ID
#     URL

import datetime
import tempfile

import luigi

from gluish.format import Gzip
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.configuration import Config
from siskin.database import sqlitedb
from siskin.task import DefaultTask

config = Config.instance()


class MAGTask(DefaultTask):
    TAG = 'mag'

    def closest(self):
        return datetime.date(2016, 2, 5)


class MAGDates(MAGTask):
    """ Extract dates. """

    date = luigi.DateParameter(default=datetime.date.today())

    def run(self):
        output = shellout("""
            curl "https://academicgraph.blob.core.windows.net/graph/index.html?eulaaccept=on" |
            grep "<li>" | grep -o "20[0-9][0-9]-[01][1-9]-[012][0-9]" | sort -u > {output} """)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())


class MAGDump(MAGTask):
    """ Dump file. """

    date = ClosestDateParameter(default=datetime.date.today())

    def run(self):
        output = shellout("""
            curl https://academicgraph.blob.core.windows.net/graph-{date}/MicrosoftAcademicGraph.zip > {output}""",
                          date=self.closest())
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="zip"))


class MAGFile(MAGTask):
    """
    Listing archive: MicrosoftAcademicGraph.zip

    --
    Path = MicrosoftAcademicGraph.zip
    Type = zip
    64-bit = +
    Physical Size = 29698421155

       Date      Time    Attr         Size   Compressed  Name
    ------------------- ----- ------------ ------------  ------------------------
    2015-11-16 15:45:20 ....A       770185       296305  Affiliations.txt
    2015-11-16 15:48:48 ....A   3037120573   1472531330  Authors.txt
    2015-11-16 15:45:22 ....A     10346759      3502468  ConferenceInstances.txt
    2015-11-16 15:45:20 ....A        79774        26235  Conferences.txt
    2015-11-16 15:45:20 ....A      1508780       713334  FieldsOfStudy.txt
    2015-11-16 15:45:22 ....A      1002313       369790  Journals.txt
    2015-06-11 22:34:30 ....A         9641         3998  license.txt
    2015-11-16 16:03:48 ....A  18017900470   4298547230  PaperAuthorAffiliations.txt
    2015-11-16 15:51:32 ....A   5329643052   1839341847  PaperKeywords.txt
    2015-11-16 16:21:14 ....A  18094921016   7421585112  PaperReferences.txt
    2015-11-16 16:27:50 ....A  27419818455   8967796102  Papers.txt
    2015-11-16 16:22:22 ....A  27259393790   5693705196  PaperUrls.txt
    2015-08-30 17:21:24 ....A         1309          484  readme.txt
    ------------------- ----- ------------ ------------  ------------------------
                               99172516117  29698419431  13 files, 0 folders
    """
    date = ClosestDateParameter(default=datetime.date.today())
    name = luigi.Parameter(default='Papers')

    def requires(self):
        return MAGDump(date=self.date)

    def run(self):
        output = shellout("7z e -so {dump} {name}.txt | pigz -c > {output}", dump=self.input().path, name=self.name)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="txt.gz"), format=Gzip)


class MAGPaperDomains(MAGTask):
    """
    Domain distribution of Paper URLs.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return MAGFile(name='PaperUrls', date=self.date)

    def run(self):
        output = shellout("""
            unpigz -c {file} | LC_ALL=C grep -o "http[s]*://[^/]*" |
            LC_ALL=C sed -e 's@http[s]*://@@g' | TMPDIR={tmpdir} LC_ALL=C sort -S50% |
            uniq -c | sort -nr > {output} """,
                          tmpdir=config.get('core', 'tempdir'), file=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="txt.gz"))


class MAGDOIList(MAGTask):
    """
    List of DOI in this dataset.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return MAGFile(date=self.date, name='Papers')

    def run(self):
        output = shellout(
            """ unpigz -c {input} | LC_ALL=C cut -f6 | LC_ALL=C grep -v ^$ | LC_ALL=C sort -S50% | pigz -c > {output} """, input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="txt.gz"))


class MAGKeywordDistribution(MAGTask):
    """
    Keyword distribution.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return MAGFile(date=self.date, name='PaperKeywords')

    def run(self):
        output = shellout("unpigz -c {input} | cut -f2 | TMPDIR={tmpdir} sort -S50% | uniq -c | sort -nr | pigz -c > {output}",
                          input=self.input().path, tmpdir=config.get('core', 'tempdir'))
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="txt.gz"))


class MAGReferenceDB(MAGTask):
    """
    Create a sqlite3 version of references. Takes about 45min, database is about 26GB.

    Referenced DOIs in 10.1002/j.2168-9830.2008.tb00991.x:

        $ sqlite3 $(taskoutput MAGReferenceDB) "select doi from lookup where id IN (\
                                                    select ref from refs where id = (\
                                                        select id from lookup where doi = '10.1002/j.2168-9830.2008.tb00991.x'))"
        10.1109/FIE.2001.964026
        10.1002/j.2168-9830.1993.tb00065.x
        10.1002/j.2168-9830.2006.tb00883.x
        10.1002/j.2168-9830.1999.tb00449.x
        10.1007/s11948-007-9040-7
        10.1002/j.2168-9830.2005.tb00823.x
        10.1002/j.2168-9830.2005.tb00836.x
        10.1177/1466138104045659

    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'papers': MAGFile(name='Papers'),
            'refs': MAGFile(name='PaperReferences'),
        }

    def run(self):
        _, dbpath = tempfile.mkstemp(prefix='siskin-')

        # Create the lookup table from doi to internal id.
        with sqlitedb(dbpath) as cursor:
            query = cursor.execute("""CREATE TABLE IF NOT EXISTS lookup (doi text, id text)""")

        batch, size = [], 1000000
        with self.input().get('papers').open() as handle:
            for i, row in enumerate(handle):
                if i % size == 0:
                    self.logger.debug("[%d] adding batch (%d) ...", i, len(batch))
                    with sqlitedb(dbpath) as cursor:
                        query = cursor.executemany("""INSERT INTO lookup VALUES (?, ?)""", batch)
                    batch = []

                fields = row.strip().split('\t')
                id, doi = fields[0].strip(), fields[5].strip()
                if id == "" or doi == "":
                    continue

                batch.append((doi, id))

        self.logger.debug("creating index...")

        with sqlitedb(dbpath) as cursor:
            query = cursor.execute("""CREATE INDEX IF NOT EXISTS idx_lookup_doi ON lookup (doi)""")
            query = cursor.execute("""CREATE INDEX IF NOT EXISTS idx_lookup_id ON lookup (id)""")

        # Create the references table from internal id to internal id.
        with sqlitedb(dbpath) as cursor:
            query = cursor.execute("""CREATE TABLE IF NOT EXISTS refs (id text, ref text)""")

        batch, size = [], 1000000
        with self.input().get('refs').open() as handle:
            for i, row in enumerate(handle):
                if i % size == 0:
                    self.logger.debug("[%d] adding batch (%d) ...", i, len(batch))
                    with sqlitedb(dbpath) as cursor:
                        query = cursor.executemany("""INSERT INTO refs VALUES (?, ?)""", batch)
                    batch = []

                fields = row.strip().split('\t')
                id, ref = fields[0].strip(), fields[1].strip()
                if id == "" or ref == "":
                    continue

                batch.append((id, ref))

        self.logger.debug("creating index...")

        with sqlitedb(dbpath) as cursor:
            query = cursor.execute("""CREATE INDEX IF NOT EXISTS idx_refs_id ON refs (id)""")

        luigi.LocalTarget(dbpath).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="db"))



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301,C0103,W0614,W0401

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Module exists, so we do not have to do star imports in utils.py.
"""

import json

from luigi.task import Register

from siskin.sources import *
from siskin.workflows import *


def _write_task_import_cache(path):
    """
    Write dict to path.
    """
    with open(path, 'w') as output:
        task_import_cache = dict([(name, Register.get_task_cls(name).__module__) for name in Register.task_names() if name[0].isupper()])
        json.dump(task_import_cache, output)



# coding: utf-8
# pylint: disable=C0301,E1101

# Copyright 2016 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Genios
------

Configuration:

[genios]

scp-src = user@ftp.example.com:/home/gbi

----

This is a reimplementation of Genios tasks.

* References (fachzeitschriften = FZS)
* Fulltexts (various packages)

* Dump (called reload, monthly)

Examples
--------

* konsortium_sachsen_fachzeitschriften_STS_reload_201609.zip
* konsortium_sachsen_fachzeitschriften_update_20160821070003.zip

Ebooks are reloaded monthy:

* konsortium_sachsen_ebooks_GABA_reload_201607.zip

* konsortium_sachsen_literaturnachweise_wirtschaftswissenschaften_IFOK_reload_201606.zip
* konsortium_sachsen_literaturnachweise_wirtschaftswissenschaften_update_20161122091046.zip

"""

import datetime
import os
import re
import tempfile
import zipfile

import luigi

from gluish.format import TSV
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.common import Executable
from siskin.sources.amsl import AMSLFilterConfig
from siskin.task import DefaultTask
from siskin.utils import iterfiles


class GeniosTask(DefaultTask):
    """
    Genios task.

    Allowed set names:

    * ebooks
    * fachzeitschriften
    * literaturnachweise_psychologie
    * literaturnachweise_recht
    * literaturnachweise_sozialwissenschaften
    * literaturnachweise_technik
    * literaturnachweise_wirtschaftswissenschaften

    The database_blacklist lists database names, that should excluded from processing, #9534.

    """
    TAG = '048'

    allowed_kinds = set([
        'ebooks',
        'fachzeitschriften',
        'literaturnachweise_psychologie',
        'literaturnachweise_recht',
        'literaturnachweise_sozialwissenschaften',
        'literaturnachweise_technik',
        'literaturnachweise_wirtschaftswissenschaften',
    ])

    database_blacklist = (
        'AE',
        'ANP',
        'AUTO',
        'AUW',
        'BSTZ',
        'EMAR',
        'FERT',
        'FLUI',
        'FP',
        'FW',
        'HOLZ',
        'INST',
        'KE',
        'KONT',
        'LEDI',
        'MEIN',
        'MF',
        'MUM',
        'NPC',
        'PROD',
        'STHZ',
        'STIF',
        'WEFO',
        'WUV',
    )

    def closest(self):
        return monthly(date=self.date)


class GeniosDropbox(GeniosTask):
    """
    Pull down content from FTP.
    """
    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return Executable('rsync', message='https://rsync.samba.org/')

    def run(self):
        target = os.path.join(self.taskdir(), 'mirror')
        shellout("mkdir -p {target} && rsync {rsync_options} {src} {target}",
                 rsync_options=self.config.get(
                     'gbi', 'rsync-options', '-avzP'),
                 src=self.config.get('gbi', 'scp-src'), target=target)

        if not os.path.exists(self.taskdir()):
            os.makedirs(self.taskdir())

        with self.output().open('w') as output:
            for path in iterfiles(target):
                output.write_tsv(path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='filelist'), format=TSV)


class GeniosReloadDates(GeniosTask):
    """
    Extract all reload dates, write them sorted into a file.

    Example output (first 4 columns):

        ...
        |  fachzeitschriften                            | ZVS     | 2016    | 10       |
        |  fachzeitschriften                            | ZWF     | 2016    | 06       |
        |  fachzeitschriften                            | ZWF     | 2016    | 07       |
        |  fachzeitschriften                            | ZWF     | 2016    | 08       |
        |  fachzeitschriften                            | ZWF     | 2016    | 09       |
        |  fachzeitschriften                            | ZWF     | 2016    | 10       |
        |  literaturnachweise_psychologie               | PSYT    | 2016    | 06       |
        |  literaturnachweise_psychologie               | PSYT    | 2016    | 07       |
        |  literaturnachweise_psychologie               | PSYT    | 2016    | 08       |
        |  literaturnachweise_psychologie               | PSYT    | 2016    | 09       |
        |  literaturnachweise_psychologie               | PSYT    | 2016    | 10       |
        ...

    Group names are hardcoded (regex) in this task. List them with:

        $ taskcat GeniosReloadDates | cut -f1|sort -u
        ebooks
        fachzeitschriften
        literaturnachweise_psychologie
        literaturnachweise_recht
        literaturnachweise_sozialwissenschaften
        literaturnachweise_technik
        literaturnachweise_wirtschaftswissenschaften

    """
    date = luigi.DateParameter(default=datetime.date.today())
    strict = luigi.BoolParameter(default=False, description='fail, if we spot bad files')

    def requires(self):
        return GeniosDropbox(date=self.date)

    def run(self):
        """
        Reload files are marked with date (years + month).
        """
        pattern = re.compile(
            '.*konsortium_sachsen_('
            'literaturnachweise_psychologie|'
            'literaturnachweise_recht|'
            'literaturnachweise_sozialwissenschaften|'
            'literaturnachweise_technik|'
            'literaturnachweise_wirtschaftswissenschaften|'
            'fachzeitschriften|ebooks)_'
            '([A-Z]*)_reload_(20[0-9][0-9])([01][0-9]).*')

        rows = []

        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('path',)):
                # If the file is too small to be zipfile, skip.
                stinfo = os.stat(row.path)
                if stinfo.st_size < 22:
                    self.logger.warning("skipping file, too small (%s): %s", stinfo.st_size, row.path)
                    if self.strict:
                        raise RuntimeError('too small: %s' % row.path)
                    continue

                # Partial Zip? Skip.
                try:
                    zipfile.ZipFile(row.path)
                except zipfile.BadZipfile as err:
                    self.logger.debug("%s: %s", err, row.path)
                    if self.strict:
                        raise err
                    continue

                match = pattern.match(row.path)
                if not match:
                    continue

                cols = list(match.groups()) + [row.path]
                # cols contains the [kind, database name, year, month, filename]
                # ['ebooks', 'DFVE', '2016', '09', '.../siskin-data/genios/...reload_201609.zip']

                if cols[1] in self.database_blacklist:
                    self.logger.debug("excluding blacklisted %s from further processing", cols[1])
                    continue

                rows.append(cols)

        with self.output().open('w') as output:
            for row in sorted(rows):
                output.write_tsv(*row)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='filelist'), format=TSV)


class GeniosDatabases(GeniosTask):
    """
    Extract a list of database names by --kind. Example output:

    AAA
    AAS
    AATG
    ABAU
    ABES
    ABIL
    ...
    """
    kind = luigi.Parameter(default='fachzeitschriften',
                           description='or: ebooks, literaturnachweise_...')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GeniosReloadDates(date=self.date)

    def run(self):
        if self.kind not in GeniosTask.allowed_kinds:
            raise RuntimeError('only these --kind parameters are allowed: %s' %
                               ', '.join(GeniosTask.allowed_kinds))

        dbs = set()
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('kind', 'db', 'year', 'month', 'path')):
                if not row.kind.startswith(self.kind):
                    continue
                dbs.add(row.db)

        with self.output().open('w') as output:
            for db in sorted(dbs):
                output.write_tsv(db)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class GeniosLatest(GeniosTask):
    """
    Get the latest version of all files, belonging to some kind, e.g. FZS.

    Latest FZS is about 10G, takes 10min to build.

    The XML contains an additional element, X-Package, containing the kind,
    e.g. fachzeitschriften, literaturnachweise_technik, ...

        ...
        </Copyright>
        <X-Package>ebooks</X-Package></Document>
        ...

    """
    kind = luigi.Parameter(default='fachzeitschriften',
                           description='or: ebooks, literaturnachweise_...')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GeniosReloadDates(date=self.date)

    def run(self):
        """
        Map each database name to the latest file. Since the required task is
        sorted, the last entry for a name should be the latest.

        Content of all files is concatenated and gzipped.

        This way, different database can have different reload dates, and we
        still can have a *latest* version without explicit dates.

        TODO: What if the latest file is a partial upload?
        """

        if self.kind not in GeniosTask.allowed_kinds:
            raise RuntimeError('only these --kind parameters are allowed: %s' %
                               ', '.join(GeniosTask.allowed_kinds))

        filemap = {}

        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('kind', 'db', 'year', 'month', 'path')):
                if not row.kind.startswith(self.kind):
                    continue
                if int(row.year) < self.date.year or int(row.month) < self.date.month:
                    self.logger.debug("maybe ignoreable, but not ignored for now (#9534): %s", row.path)
                filemap[row.db] = row.path

        if not filemap:
            raise RuntimeError(
                'could not file a single file for the specified kind: %s' % self.kind)

        _, stopover = tempfile.mkstemp(prefix='siskin-')
        for _, path in filemap.items():
            shellout(r"""unzip -p {input} |
                        iconv -f iso-8859-1 -t utf-8 |
                        LC_ALL=C grep -v "^<\!DOCTYPE GENIOS PUBLIC" |
                        LC_ALL=C sed -e 's@<?xml version="1.0" encoding="ISO-8859-1" ?>@@g' |
                        LC_ALL=C sed -e 's@</Document>@<X-Package>{package}</X-Package></Document>@' |
                        pigz -c >> {output}""", input=path, output=stopover, package=self.kind)

        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='xml.gz'))


class GeniosIntermediateSchema(GeniosTask):
    """
    Intermediate schema by kind. May be incomplete, since the database mapping
    is derived from dozens of XLS sheets and manual guesses.

    Related: "Neue Quellen bzw. Austausch", Mon, Dec 5, 2016 at 12:23 PM, ba54ea7d396a41a2a1281f51bba5d33f
    See also: #9534.
    """
    kind = luigi.Parameter(default='fachzeitschriften',
                           description='or: ebooks, literaturnachweise_...')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return GeniosLatest(kind=self.kind, date=self.date)

    def run(self):
        if not os.path.exists(self.taskdir()):
            os.makedirs(self.taskdir())
        output = shellout("span-import -i genios <(unpigz -c {input}) | pigz -c >> {output}", input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'))


class GeniosCombinedIntermediateSchema(GeniosTask):
    """
    Concat all genios files.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        """
        No ebooks.
        """
        return [
            GeniosIntermediateSchema(date=self.date, kind='fachzeitschriften'),
            GeniosIntermediateSchema(
                date=self.date, kind='literaturnachweise_psychologie'),
            GeniosIntermediateSchema(
                date=self.date, kind='literaturnachweise_recht'),
            GeniosIntermediateSchema(
                date=self.date, kind='literaturnachweise_sozialwissenschaften'),
            GeniosIntermediateSchema(
                date=self.date, kind='literaturnachweise_technik'),
            GeniosIntermediateSchema(
                date=self.date, kind='literaturnachweise_wirtschaftswissenschaften'),
        ]

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        for target in self.input():
            shellout("cat {input} >> {output}",
                     input=target.path, output=stopover)
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'))


class GeniosCombinedExport(GeniosTask):
    """
    Tag with ISILs, then export to various formats.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    format = luigi.Parameter(default='solr5vu3', description='export format')

    def requires(self):
        return {
            'file': GeniosCombinedIntermediateSchema(date=self.date),
            'config': AMSLFilterConfig(date=self.date),
        }

    def run(self):
        output = shellout("span-tag -c {config} <(unpigz -c {input}) | pigz -c > {output}",
                          config=self.input().get('config').path, input=self.input().get('file').path)
        output = shellout(
            "span-export -o {format} <(unpigz -c {input}) | pigz -c > {output}", format=self.format, input=output)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        extensions = {
            'solr5vu3': 'ldj.gz',
            'formeta': 'form.gz',
        }
        return luigi.LocalTarget(path=self.path(ext=extensions.get(self.format, 'gz')))


class GeniosISSNList(GeniosTask):
    """
    A list of Genios ISSNs.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'input': GeniosCombinedIntermediateSchema(date=self.date),
            'jq': Executable(name='jq', message='https://github.com/stedolan/jq')
        }

    def run(self):
        _, output = tempfile.mkstemp(prefix='siskin-')
        shellout("""jq -c -r '.["rft.issn"][]?' <(unpigz -c {input}) >> {output} """, input=self.input().get('input').path, output=output)
        shellout("""jq -c -r '.["rft.eissn"][]?' <(unpigz -c {input}) >> {output} """, input=self.input().get('input').path, output=output)
        output = shellout("""LC_ALL=C sort -S35% -u {input} > {output} """, input=output)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)



# coding: utf-8
# pylint: disable=C0301,E1101

# Copyright 2016 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
BASE.

* https://www.base-search.net/
* https://api.base-search.net/
* https://en.wikipedia.org/wiki/BASE_(search_engine)
* http://www.dlib.org/dlib/june04/lossau/06lossau.html

"""

import datetime

import luigi

from gluish.intervals import weekly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.common import Executable
from siskin.task import DefaultTask


class BaseTask(DefaultTask):
    """
    Various tasks around base.
    """
    TAG = 'base'

    def closest(self):
        return weekly(self.date)


class BaseDynamicSet(BaseTask):
    """
    TODO(miku): Dynamic sets use SOLR queries:

    set=classcode:(002 OR 070 OR 659 OR 791)     => fq on SORL field classcode
    set=subject:(*journalismus* OR *journalism*) => fq on SOLR field subject
    set=classcode:(002 OR 070 OR 659 OR 791) subject:(*journalismus* OR *journalism*) => fq on SOLR field classcode and subject

    Long set names currently cause problems with metha-sync.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    prefix = luigi.Parameter(default="oai_dc", significant=False)
    set = luigi.Parameter(description='a subject / keyword')

    def requires(self):
        return Executable(name='metha-sync', message='https://github.com/miku/metha')

    def run(self):
        shellout("METHA_DIR={dir} metha-sync -set '{set}' -format {prefix} http://oai.base-search.net/oai",
                 set=self.set, prefix=self.prefix, dir=self.config.get('core', 'metha-dir'))
        output = shellout("METHA_DIR={dir} metha-cat -set '{set}' -format {prefix} http://oai.base-search.net/oai | pigz -c > {output}",
                          set=self.set, prefix=self.prefix, dir=self.config.get('core', 'metha-dir'))
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="xml.gz", digest=True))



# coding: utf-8
# pylint: disable=C0301,C0103

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Helper for databases.
"""

from future import standard_library
standard_library.install_aliases()
from builtins import object
import logging
import sqlite3
import urllib.parse

import pymysql
from pymysql.cursors import SSCursor


class sqlitedb(object):
    """
    Simple context manager for sqlite3 databases. Commits everything at exit.

        with sqlitedb('/tmp/test.db') as cursor:
            query = cursor.execute('SELECT * FROM items')
            result = query.fetchall()
    """

    def __init__(self, path, timeout=5.0, detect_types=0):
        self.path = path
        self.conn = None
        self.cursor = None
        self.timeout = timeout
        self.detect_types = detect_types

    def __enter__(self):
        self.conn = sqlite3.connect(
            self.path, timeout=self.timeout, detect_types=self.detect_types)
        self.conn.text_factory = str
        self.cursor = self.conn.cursor()
        return self.cursor

    def __exit__(self, exc_class, exc, traceback):
        self.conn.commit()
        self.conn.close()


class mysqldb(object):
    """ Context manager for MySQL database access.

        with mysqldb('mysql://user:pass@host/db', stream=True) as cursor:
            query = cursor.execute('SELECT * FROM items')
            result = query.fetchall()

    """

    def __init__(self, url, stream=False, commit_on_exit=False):
        result = urllib.parse.urlparse(url, scheme='mysql')
        self.hostname = result.hostname
        self.username = result.username
        self.password = result.password
        self.database = result.path.strip('/')
        self.stream = stream
        self.commit_on_exit = commit_on_exit
        self.conn = None
        self.cursor = None

    def __enter__(self):
        if self.stream:
            self.conn = pymysql.connect(host=self.hostname, user=self.username,
                                        passwd=self.password, db=self.database,
                                        cursorclass=SSCursor)
        else:
            self.conn = pymysql.connect(host=self.hostname, user=self.username,
                                        passwd=self.password, db=self.database)

        self.cursor = self.conn.cursor()
        return self.cursor

    def __exit__(self, exc_class, exc, traceback):
        if self.commit_on_exit:
            self.conn.commit()
        self.cursor.close()
        self.conn.close()



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Robert Schenk, <robert.schenk@uni-leipzig.de>
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
VKFilmBerlin, UdK Berlin, #8697.
"""

import datetime

import luigi

from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.sources.b3kat import B3KatDownload
from siskin.task import DefaultTask


class VKFilmBerlinTask(DefaultTask):
    TAG = '117'

    def closest(self):
        return monthly(date=self.date)


class VKFilmBerlinFiltered(VKFilmBerlinTask):
    """
    Apply filter.
    """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return B3KatDownload(date=self.date)

    def run(self):
        output = shellout("flux.sh {flux} in={input} > {output}",
                          flux=self.assets("117/flux_b3kat_filter.flux"), input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='xml'))


class VKFilmBerlinMARC(VKFilmBerlinTask):
    """
    Apply transformations.
    """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return VKFilmBerlinFiltered(date=self.date)

    def run(self):
        """ https://stackoverflow.com/a/7774512 """
        output = shellout(r"""
            perl -CSDA -pe's/[^\x9\xA\xD\x20-\x{{D7FF}}\x{{E000}}-\x{{FFFD}}\x{{10000}}-\x{{10FFFF}}]+//g;' {input} > {output}
        """, input=self.input().path)
        output = shellout("flux.sh {flux} in={input} > {output}",
                          flux=self.assets("117/117.flux"), input=output)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='fincmarc.xml'))



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
IEEE. (http://www.ieee.org/documents/xmldoc_ver5.5.pdf)

Config
------

[ieee]

ftp-host = host.name
ftp-username = username
ftp-password = password
ftp-path = /
ftp-pattern = *

# initial shipment compressed as tar.gz
backlog-archive = /path/to/ieee.tar.gz

"""

import datetime
import tempfile

import luigi

from gluish.common import Executable
from gluish.format import TSV
from gluish.intervals import weekly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.benchmark import timed
from siskin.common import FTPMirror
from siskin.task import DefaultTask


class IEEETask(DefaultTask):
    TAG = 'ieee'

    def closest(self):
        return weekly(date=self.date)


class IEEEPaths(IEEETask):
    """ A list of IEEE file paths (via FTP). """
    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        host = self.config.get('ieee', 'ftp-host')
        username = self.config.get('ieee', 'ftp-username')
        password = self.config.get('ieee', 'ftp-password')
        base = self.config.get('ieee', 'ftp-path')
        pattern = self.config.get('ieee', 'ftp-pattern')
        return FTPMirror(host=host, username=username, password=password,
                         base=base, pattern=pattern)

    @timed
    def run(self):
        output = shellout('sort {input} > {output}', input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="filelist"), format=TSV)


class IEEEUpdatesIntermediateSchema(IEEETask):
    """
    Combined updates, with duplicates.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return IEEEPaths(date=self.date)

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('path',)):
                if not 'IEEEUpdates_' in row.path:
                    continue
                shellout(r'unzip -p {input} \*.xml | span-import -i ieee | pigz -c >> {output}', input=row.path, output=stopover)
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="ldj.gz"))


class IEEEBacklogPaths(IEEETask):
    """
    List files in the backlog. Just a `tar -tf` of the compressed dump.
    """

    def run(self):
        output = shellout('tar -tf {input} > {output}', input=self.config.get('ieee', 'backlog-archive'))
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="filelist"), format=TSV)


class IEEEBacklogIntermediateSchema(IEEETask):
    """
    Backlog is a single tar file containing an initial set of records.
    """

    def run(self):
        output = shellout("""tar xOzf {input} | span-import -i ieee | pigz -c >> {output}""",
                          input=self.config.get('ieee', 'backlog-archive'), ignoremap={2: 'FIXME later'})
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True, ext='ldj.gz'))


class IEEEIntermediateSchema(IEEETask):
    """
    Combine the backlog and all updates into a single file.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return [IEEEBacklogIntermediateSchema(), IEEEUpdatesIntermediateSchema(date=self.date)]

    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        for target in self.input():
            shellout("cat {input} >> {output}", input=target.path, output=stopover)
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="ldj.gz"))


class IEEESolrExport(IEEETask):
    """
    Export to a SOLR compatible format.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return IEEEIntermediateSchema(date=self.date)

    def run(self):
        """
        TODO: get ISIL attachments from AMSL.
        """
        output = shellout("""span-tag -c '{{"DE-15": {{"any": {{}} }} }}' <(unpigz -c {input}) | span-export | pigz -c > {output}""", input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="ldj.gz"))


class IEEEDOIList(IEEETask):
    """ A list of IEEE DOIs. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'input': IEEEIntermediateSchema(date=self.date),
                'jq': Executable(name='jq', message='https://github.com/stedolan/jq')}

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        shellout("""jq -r '.doi' <(unpigz -c {input}) | grep -v "null" | grep -o "10.*" 2> /dev/null > {output} """,
                 input=self.input().get('input').path, output=stopover)
        output = shellout("""sort -u {input} > {output} """, input=stopover)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301
# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Tracy Hoffmann, <tracy.hoffmann@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
IZI Task #7755
"""

import datetime

import luigi

from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.task import DefaultTask

class IZITask(DefaultTask):
    """ Base task for IZI """
    TAG = '78'

class IZIIntermediateSchema(IZITask):
    """ Convert XML to intermediate schema via metafacture """
    date = ClosestDateParameter(default=datetime.date.today())
    iziFile = luigi.Parameter(default='/tmp/IZI_XML.xml', description='path izi datadump', significant=False)

    def run(self):
        output = shellout("""flux.sh {flux} in={iziFile} > {output}""",
                         flux=self.assets("78/78.flux"), iziFile=self.iziFile)
        luigi.LocalTarget(output).move(self.output().path)
    
    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))

class IZIFincSolr(IZITask):
    """
    Export to finc solr schema by using span-export.
    Tag with ISIL for FID and change record type.
    """
    format = luigi.Parameter(default='solr5vu3', description='export format')
    isil = luigi.Parameter(default='DE-15-FID', description='isil FID')
    date = ClosestDateParameter(default=datetime.date.today())
    iziFile = luigi.Parameter(default='/tmp/IZI_XML.xml', description='path izi datadump', significant=False)

    def requires(self):
        return {
            'file': IZIIntermediateSchema(date=self.date, iziFile=self.iziFile)
        }
    
    def run(self):
        output = shellout("""span-export -o {format} -with-fullrecord <(span-tag -c <(echo '{{"{isil}": {{"any": {{}}}}}}') {input}) > {output}""",
                          format=self.format, isil=self.isil, input=self.input().get('file').path)
        output = shellout(
            """cat {input} | sed 's/"recordtype":"ai"/"recordtype":"is"/g' > {output}""", input=output)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='fincsolr.ndj'))


# coding: utf-8
# pylint: disable=C0301,E1101

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Zenodo test task for generic OAI.
"""

import datetime

import luigi

from siskin.oai import OAIIntermediateSchema
from siskin.task import DefaultTask


class ZenodoTask(DefaultTask):
    TAG = 'zenodo'


class ZenodoIntermediateSchema(ZenodoTask, luigi.WrapperTask):
    """ Convert Zenodo to intermediate schema. """
    date = luigi.DateParameter(default=datetime.date.today())
    url = luigi.Parameter(default="https://zenodo.org/oai2d", significant=False)
    prefix = luigi.Parameter(default="oai_dc", significant=False)

    sid = luigi.Parameter(default='zenodo', description="internal data source id")
    collection = luigi.Parameter(default='Zenodo', description="internal collection name")
    format = luigi.Parameter(default='ElectronicArticle')

    def requires(self):
        return OAIIntermediateSchema(date=self.date, url=self.url, prefix=self.prefix, sid='zenodo', collection='Zenodo', format=self.format)

    def output(self):
        return self.input()



# coding: utf-8
# pylint: disable=C0301,E1101

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
https://www.ceeol.com/, refs #9398.

----

Config:

[ceeol]

disk-dir = /path/to/disk


"""

from __future__ import print_function

import glob
import os
import tempfile

import luigi
from gluish.utils import shellout
from siskin.task import DefaultTask


class CeeolTask(DefaultTask):
    """
    Base task.
    """
    TAG = '53'


class CeeolJournalsDumpIntermediateSchema(CeeolTask):
    """
    Combine all journals from disk dump and convert them to intermediate schema.
    """

    def run(self):
        diskdir = self.config.get('ceeol', 'disk-dir')
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        for path in glob.glob(os.path.join(diskdir, '1Journal-Articles-XML', 'articles_*xml')):
            shellout("span-import -i ceeol {input} | pigz -c >> {output}", input=path, output=stopover)
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="ldj.gz"))



# coding: utf-8
# pylint: disable=C0301,R0904,W0221

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Configuration handling, taken from luigi. File format is ini. File location
at the moment is fixed at /etc/siskin/siskin.ini.
"""

import datetime
import logging
import os
import sys
from ConfigParser import ConfigParser, NoOptionError, NoSectionError

logger = logging.getLogger('siskin')


class Config(ConfigParser):
    """ Wrapper around /etc/siskin/siskin.ini
    """
    NO_DEFAULT = None
    _instance = None

    # most specific path last
    _config_paths = [
        '/etc/siskin/siskin.ini',
        os.path.join(os.path.expanduser('~'), '.config/siskin/siskin.ini'),
    ]

    @classmethod
    def add_config_path(cls, path):
        """ Append config path. """
        cls._config_paths.append(path)
        cls._instance.reload()

    @classmethod
    def instance(cls, *args, **kwargs):
        """ Singleton getter """
        if cls._instance is None:
            cls._instance = cls(*args, **kwargs)
            _ = cls._instance.reload()

        return cls._instance

    def reload(self):
        """ Reload configuration. """
        return self._instance.read(self._config_paths)

    def _get_with_default(self, method, section, option, default, expected_type=None):
        """ Gets the value of the section/option using method. Returns default if value
        is not found. Raises an exception if the default value is not None and doesn't match
        the expected_type.
        """
        try:
            return method(self, section, option)
        except (NoOptionError, NoSectionError) as err:
            if default is Config.NO_DEFAULT:
                logger.error('invalid or missing configuration: %s', err)
                raise
            if expected_type is not None and default is not None and not isinstance(default, expected_type):
                logger.error('invalid or missing configuration: %s', err)
                raise
            return default

    def get(self, section, option, default=NO_DEFAULT):
        return self._get_with_default(ConfigParser.get, section, option, default)

    def getboolean(self, section, option, default=NO_DEFAULT):
        return self._get_with_default(ConfigParser.getboolean, section, option, default, bool)

    def getint(self, section, option, default=NO_DEFAULT):
        return self._get_with_default(ConfigParser.getint, section, option, default, int)

    def getfloat(self, section, option, default=NO_DEFAULT):
        return self._get_with_default(ConfigParser.getfloat, section, option, default, float)

    def getdate(self, section, option, default=NO_DEFAULT):
        """ Dates must be formatted ISO8601 style: %Y-%m-%d """
        value = self.get(section, option, default=default)
        return datetime.date(*map(int, value.split('-')))



# coding: utf-8
# pylint: disable=C0103,W0232,C0301

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Define a siskin wide task with artefacts under core.home directory.

[amsl]

write-url = https://live.abc.xyz/w/i/write

"""

import logging

import os
import re
import tempfile

import luigi

from gluish.task import BaseTask
from gluish.utils import shellout
from siskin.configuration import Config

config = Config.instance()


class DefaultTask(BaseTask):
    """ A base task that sets its base directory based on config value. """
    BASE = config.get('core', 'home', os.path.join(tempfile.gettempdir(), 'siskin-data'))

    stamp = luigi.BoolParameter(default=False, description="send an updated stamp to AMSL", significant=False)

    def assets(self, path):
        """ Return the absolute path to the asset. `path` is the relative path
        below the assets root dir. """
        return os.path.join(os.path.dirname(__file__), 'assets', path)

    @property
    def config(self):
        """
        Return the config instance.
        """
        return config

    @property
    def logger(self):
        """
        Return the logger. Module logging uses singleton internally, so no worries.
        """
        return logging.getLogger('siskin')

    def on_success(self):
        """
        Try to send a datestamp to AMSL, but only if a couple of prerequisites are met:

        All subclasses inherit a --stamp boolean flag, which must be set. If
        the TAG of the source is not numeric, we won't do anything. If
        "amsl.write-url" configuration is not set, we will log the error, but
        do not stop processing. Finally, even if the HTTP request fails, it
        won't be fatal.

        On success, the API returns:

            < HTTP/1.1 200 OK
            < Content-Type: text/html; charset=UTF-8
            < Content-Length: 2
            ...

            OK
        """

        if not self.stamp:
            return

        if not hasattr(self, 'TAG'):
            self.logger.warn("No tag defined, skip stamping.")
            return

        if not re.match(r"^[\d]+$", self.TAG):
            self.logger.warn("Non-integer source id: %s, skip stamping.", self.TAG)
            return

        # Otherwise: Parameter 'sid' ... not a positive integer.
        sid = self.TAG.lstrip("0")

        try:
            write_url = config.get("amsl", "write-url", None)
            if write_url is None:
                self.logger.warn("Missing amsl.write-url configuration, skip stamping.")
                return
        except Exception as err:
            self.logger.warn("Could not stamp: %s", err)
            return

        try:
            shellout("""curl --fail -XPOST "{write_url}?do=updatetime&sid={sid}" > /dev/null """, write_url=write_url, sid=sid)
        except RuntimeError as err:
            self.logger.warn(err)
            return
        else:
            self.logger.debug("Successfully stamped: %s", sid)



# coding: utf-8
# pylint: disable=C0301

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

#
# Adhoc
# =====
#

import luigi

from gluish.format import TSV
from gluish.utils import shellout
from siskin.sources.crossref import CrossrefExport
from siskin.sources.degruyter import DegruyterExport
from siskin.sources.doaj import DOAJExport
from siskin.sources.elsevierjournals import ElsevierJournalsExport
from siskin.sources.highwire import HighwireExport
from siskin.sources.jstor import JstorExport
from siskin.task import DefaultTask


class AdhocTask(DefaultTask):
    """
    Base task for throwaway tasks.
    """
    TAG = "adhoc"


class AdhocFormetaSamples(AdhocTask):
    """
    Formeta test samples.
    """

    def requires(self):
        return [
            CrossrefExport(format='formeta'),
            DegruyterExport(format='formeta'),
            DOAJExport(format='formeta'),
            ElsevierJournalsExport(format='formeta'),
            HighwireExport(format='formeta'),
            JstorExport(format='formeta'),
        ]

    def run(self):
        with self.output().open('w') as output:
            for target in self.input():
                output.write_tsv(target.path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Highwire, containing data from (approximate counts):

1764237 Oxford University Press
 558404 BMJ Publishing Group Ltd
  42011 Company of Biologists
  25004 Duke University Press
  15971 Cold Spring Harbor Laboratory Press
   5945 British Medical Journal Publishing Group
   1212 The Company of Biologists Ltd
    149 eLife Sciences Publications Limited
     81 The Company of Biologists

"""

import datetime

import luigi

from gluish.common import Executable
from gluish.intervals import weekly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.task import DefaultTask


class HighwireTask(DefaultTask):
    """
    Base task.
    """
    TAG = 'highwire'

    def closest(self):
        return weekly(date=self.date)


class HighwireCombine(HighwireTask):
    """
    Single file dump.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    url = luigi.Parameter(default="open-archive.highwire.org/handler", significant=False)
    prefix = luigi.Parameter(default="oai_dc", significant=False)

    def requires(self):
        return Executable(name='metha-sync', message='https://github.com/miku/metha')

    def run(self):
        shellout("METHA_DIR={dir} metha-sync -ignore-http-errors -format {prefix} {url}",
                 prefix=self.prefix, url=self.url, dir=self.config.get('core', 'metha-dir'))
        output = shellout("METHA_DIR={dir} metha-cat -format {prefix} {url} | pigz -c > {output}",
                          prefix=self.prefix, url=self.url, dir=self.config.get('core', 'metha-dir'))
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="xml.gz"))


class HighwireIntermediateSchema(HighwireTask):
    """
    Experimental OAI -> intermediate schema conversion.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    url = luigi.Parameter(default="open-archive.highwire.org/handler", significant=False)
    prefix = luigi.Parameter(default="oai_dc", significant=False)

    def requires(self):
        return HighwireCombine(date=self.date, url=self.url, prefix=self.prefix)

    def run(self):
        output = shellout("unpigz -c {input} | span-import -i highwire | pigz -c > {output}",
                          input=self.input().path, pipefail=False)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="ldj.gz"))


class HighwireExport(HighwireTask):
    """
    Export to various formats
    """
    date = ClosestDateParameter(default=datetime.date.today())
    format = luigi.Parameter(default='solr5vu3', description='export format')

    def requires(self):
        return HighwireIntermediateSchema(date=self.date)

    def run(self):
        output = shellout("span-export -o {format} <(unpigz -c {input}) | pigz -c > {output}", format=self.format, input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        extensions = {
            'solr5vu3': 'ldj.gz',
            'formeta': 'form.gz',
        }
        return luigi.LocalTarget(path=self.path(ext=extensions.get(self.format, 'gz')))



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
[jstor]

ftp-host = host.name
ftp-username = username
ftp-password = password
ftp-path = /
ftp-pattern = *
"""

import datetime
import itertools
import pipes
import tempfile

import luigi

from gluish.format import TSV
from gluish.intervals import weekly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.benchmark import timed
from siskin.common import Executable, FTPMirror
from siskin.sources.amsl import AMSLFilterConfig
from siskin.task import DefaultTask
from siskin.utils import nwise


class JstorTask(DefaultTask):
    """ Jstor base. """
    TAG = 'jstor'

    def closest(self):
        return weekly(self.date)


class JstorPaths(JstorTask):
    """
    Sync.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    max_retries = luigi.IntParameter(default=10, significant=False)
    timeout = luigi.IntParameter(
        default=20, significant=False, description='timeout in seconds')

    def requires(self):
        return FTPMirror(host=self.config.get('jstor', 'ftp-host'),
                         username=self.config.get('jstor', 'ftp-username'),
                         password=self.config.get('jstor', 'ftp-password'),
                         pattern=self.config.get('jstor', 'ftp-pattern'),
                         max_retries=self.max_retries,
                         timeout=self.timeout)

    def run(self):
        self.input().move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class JstorMembers(JstorTask):
    """
    Extract a full list of archive members.
    TODO: This should only be done once per file.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return JstorPaths(date=self.date)

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('path',)):
                if not row.path.endswith('.zip'):
                    self.logger.debug('skipping: %s', row.path)
                    continue
                shellout(""" unzip -l {input} | LC_ALL=C grep "xml$" | LC_ALL=C awk '{{print "{input}\t"$4}}' | LC_ALL=C sort >> {output} """,
                         preserve_whitespace=True, input=row.path, output=stopover)
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class JstorLatestMembers(JstorTask):
    """
    Find the latest archive members for all article ids.

    The path names are like XXX_YYYY-MM-DD-HH-MM-SS/jjjj/yyyy-mm-dd/id/id.xml.

    This way, it is possible to sort the shipments by date, run `tac` and
    only keep the latest entries.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return JstorMembers(date=self.date)

    @timed
    def run(self):
        """
        Expect input to be sorted by shipment date, so tac will actually be a perfect rewind.
        """
        output = shellout("tac {input} | LC_ALL=C sort -S 35% -u -k2,2 | LC_ALL=C sort -S 35% -k1,1 > {output}",
                          input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class JstorXML(JstorTask):
    """
    Create a snapshot of the latest data.
    TODO(miku): maybe shard by journal and reduce update time.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    batch = luigi.IntParameter(default=512, significant=False)

    def requires(self):
        return JstorLatestMembers(date=self.date)

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        with self.input().open() as handle:
            groups = itertools.groupby(handle.iter_tsv(
                cols=('archive', 'member')), lambda row: row.archive)
            for archive, items in groups:
                for chunk in nwise(items, n=self.batch):
                    margs = " ".join(["'%s'" % item.member.replace(
                        '[', r'\[').replace(']', r'\]') for item in chunk])
                    shellout("""unzip -p {archive} {members} |
                                sed -e 's@<?xml version="1.0" encoding="UTF-8"?>@@g' | pigz -c >> {output}""",
                             archive=archive, members=margs, output=stopover)

        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='xml.gz'), format=TSV)


class JstorIntermediateSchema(JstorTask):
    """
    Convert to intermediate format via span.
    """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'span': Executable(name='span-import', message='http://git.io/vI8NV'),
            'file': JstorXML(date=self.date)
        }

    @timed
    def run(self):
        output = shellout("span-import -i jstor <(unpigz -c {input}) | pigz -c > {output}",
                          input=self.input().get('file').path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'))


class JstorExport(JstorTask):
    """
    Tag with ISILs, then export to various formats.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    format = luigi.Parameter(default='solr5vu3', description='export format')

    def requires(self):
        return {
            'file': JstorIntermediateSchema(date=self.date),
            'config': AMSLFilterConfig(date=self.date),
        }

    def run(self):
        output = shellout("span-tag -c {config} <(unpigz -c {input}) | pigz -c > {output}",
                          config=self.input().get('config').path, input=self.input().get('file').path)
        output = shellout(
            "span-export -o {format} <(unpigz -c {input}) | pigz -c > {output}", format=self.format, input=output)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        extensions = {
            'solr5vu3': 'ldj.gz',
            'formeta': 'form.gz',
        }
        return luigi.LocalTarget(path=self.path(ext=extensions.get(self.format, 'gz')))


class JstorISSNList(JstorTask):
    """
    A list of JSTOR ISSNs.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return JstorIntermediateSchema(date=self.date)

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        shellout(
            """jq -r '.["rft.issn"][]?' <(unpigz -c {input}) 2> /dev/null >> {output} """, input=self.input().path, output=stopover)
        shellout(
            """jq -r '.["rft.eissn"][]?' <(unpigz -c {input}) 2> /dev/null >> {output} """, input=self.input().path, output=stopover)
        output = shellout("""sort -u {input} > {output} """, input=stopover)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class JstorDOIList(JstorTask):
    """
    A list of JSTOR DOIs.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return JstorIntermediateSchema(date=self.date)

    @timed
    def run(self):
        output = shellout("""jq -r '.doi' <(unpigz -c {input}) | grep -v null > {output} """,
                          input=self.input().path)
        output = shellout("""sort -u {input} > {output} """, input=output)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Springer (testing).

Config
------

[springer]

intermediate-schema-file = /path/to/file # must be gzip compressed

username = nana
password = s3cret
url = http://export.com/intermediate.file.gz
"""

from __future__ import print_function

import datetime
import json
import re

import luigi

from gluish.format import Gzip
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.sources.amsl import AMSLFilterConfig
from siskin.task import DefaultTask


class SpringerTask(DefaultTask):
    TAG = 'springer'

    def closest(self):
        return monthly(date=self.date)


class SpringerProvided(SpringerTask, luigi.ExternalTask):
    """
    Provided.
    """

    def output(self):
        return luigi.LocalTarget(path=self.config.get('springer', 'intermediate-schema-file'), format=Gzip)


class SpringerDownload(SpringerTask):
    """
    Attempt to download file from a configured URL.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def run(self):
        output = shellout(""" curl --fail -v -u {username}:{password} "{url}" > {output} """,
                          username=self.config.get('springer', 'username'),
                          password=self.config.get('springer', 'password'),
                          url=self.config.get('springer', 'url'))
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)


class SpringerCleanFields(SpringerTask):
    """
    Clean abstracts, refs #...
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return SpringerDownload(date=self.date)

    def run(self):
        striptags = lambda s: re.sub(r'\$\$[^\$]*\$\$', '', re.sub(r'<[^>]*>', '', s))
        with self.input().open() as handle:
            with self.output().open('w') as output:
                for line in handle:
                    doc = json.loads(line)
                    doc['abstract'] = striptags(doc.get('abstract', '').encode('utf-8'))
                    doc['rft.atitle'] = striptags(doc.get('rft.atitle', '').encode('utf-8'))
                    doc['x.subjects'] = [striptags(subj.encode('utf-8')) for subj in doc.get('x.subjects', [])]
                    output.write(json.dumps(doc))
                    output.write("\n")

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'), format=Gzip)


class SpringerIntermediateSchema(SpringerTask, luigi.WrapperTask):
    """
    Just the cleaned version.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return SpringerCleanFields(date=self.date)

    def output(self):
        return self.input()



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301,C0330

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

import datetime

import luigi

from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.task import DefaultTask

"""
Example workflow with OAI harvest and metafacture.
"""


class IJOCTask(DefaultTask):
    """ Base task for International Journal of Communication """
    TAG = '87'

    def closest(self):
        return monthly(date=self.date)


class IJOCHarvest(IJOCTask):
    """
    Harvest.
    """
    endpoint = luigi.Parameter(
        default='http://ijoc.org/index.php/ijoc/oai', significant=False)
    date = ClosestDateParameter(default=datetime.date.today())

    def run(self):
        shellout("""metha-sync "{endpoint}" """, endpoint=self.endpoint)
        output = shellout(
            """metha-cat -root Records "{endpoint}" > {output}""", endpoint=self.endpoint)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='xml'))


class IJOCIntermediateSchema(IJOCTask):
    """
    Convert to intermediate schema via metafacture. Custom morphs and flux are kept in assets/87.
    Maps are kept in assets/maps
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return IJOCHarvest(date=self.date)

    def run(self):
        mapdir = 'file:///%s' % self.assets("maps/")
        output = shellout("""flux.sh {flux} in={input} MAP_DIR={mapdir} > {output}""",
                          flux=self.assets("87/87.flux"), mapdir=mapdir, input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))


class IJOCFincSolr(IJOCTask):
    """
    Export to finc solr schema by using span-export.
    Tag with ISIL for FID and change record type.
    """
    format = luigi.Parameter(default='solr5vu3', description='export format')
    isil = luigi.Parameter(default='DE-15-FID', description='isil FID')
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'file': IJOCIntermediateSchema(date=self.date)
        }

    def run(self):
        output = shellout("""span-export -o {format} -with-fullrecord <(span-tag -c <(echo '{{"{isil}": {{"any": {{}}}}}}') {input}) > {output}""",
                          format=self.format, isil=self.isil, input=self.input().get('file').path)
        output = shellout(
            """cat {input} | sed 's/"recordtype":"ai"/"recordtype":"is"/g' > {output}""", input=output)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='fincsolr.ndj'))



# coding: utf-8
# pylint: disable=F0401,W0232,E1101,C0103,C0301,W0223,E1123,R0904,E1103

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Directory of Open Access Journals.

DOAJ is an online directory that indexes and provides access to high quality,
open access, peer-reviewed journals.

http://doaj.org
"""

from builtins import map
from builtins import range
import datetime
import itertools
import json
import operator
import tempfile
import time

import luigi

import elasticsearch
from gluish.common import Executable
from gluish.format import TSV
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.benchmark import timed
from siskin.task import DefaultTask
from siskin.utils import load_set_from_file, load_set_from_target


class DOAJTask(DefaultTask):
    """
    Base task for DOAJ.
    """
    TAG = '028'

    def closest(self):
        return monthly(date=self.date)


class DOAJCSV(DOAJTask):
    """
    CSV dump, updated every 30 minutes. Not sure what's in there.
    """
    date = luigi.DateParameter(default=datetime.date.today())
    url = luigi.Parameter(default='http://doaj.org/csv', significant=False)

    def requires(self):
        return Executable(name='wget', message='http://www.gnu.org/software/wget/')

    @timed
    def run(self):
        output = shellout('wget --retry-connrefused {url} -O {output}', url=self.url)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='csv'))


class DOAJDump(DOAJTask):
    """
    Complete DOAJ Elasticsearch dump., refs: #2089.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    host = luigi.Parameter(default='doaj.org', significant=False)
    port = luigi.IntParameter(default=443, significant=False)
    url_prefix = luigi.Parameter(default='query', significant=False)

    batch_size = luigi.IntParameter(default=1000, significant=False)
    timeout = luigi.IntParameter(default=60, significant=False)
    max_retries = luigi.IntParameter(default=3, significant=False)

    @timed
    def run(self):
        """
        Connect to ES and issue queries. Use exponential backoff to mitigate
        gateway timeouts. Be light on resources and do not crawl in parallel.
        """
        max_backoff_retry = 10
        backoff_interval_s = 0.05

        hosts = [{'host': self.host, 'port': self.port, 'url_prefix': self.url_prefix}]
        es = elasticsearch.Elasticsearch(hosts, timeout=self.timeout, max_retries=self.max_retries, use_ssl=True)
        with self.output().open('w') as output:
            offset, total = 0, 0
            while offset <= total:
                for i in range(1, max_backoff_retry + 1):
                    self.logger.debug(json.dumps({'attempt': i, 'offset': offset, 'total': total}))

                    try:
                        result = es.search(body={'constant_score': {'query': {'match_all': {}}}},
                                           index=('journal', 'article'),
                                           size=self.batch_size, from_=offset)
                    except Exception:
                        if i == max_backoff_retry:
                            raise
                        time.sleep(backoff_interval_s)
                        backoff_interval_s = 2 * backoff_interval_s
                        continue

                    for doc in result['hits']['hits']:
                        output.write("%s\n" % json.dumps(doc))
                    total = total or result['hits']['total']
                    offset += self.batch_size
                    break

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))


class DOAJIdentifierBlacklist(DOAJTask):
    """
    Create a blacklist of identifiers.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    min_title_length = luigi.IntParameter(default=30,
                                          description='Only consider titles with at least this length.')

    def requires(self):
        return DOAJDump(date=self.date)

    def run(self):
        """
        Output a list of identifiers of documents, which have the same title. Use the newest.
        """
        output = shellout("""jq -rc '[
                                .["_source"]["bibjson"]["title"],
                                .["_source"]["last_updated"],
                                .["_source"]["id"]
                            ]' {input} | sort -S35% > {output}""", input=self.input().path)

        with open(output) as handle:
            docs = (json.loads(s) for s in handle)

            with self.output().open('w') as output:
                for title, grouper in itertools.groupby(docs, key=operator.itemgetter(0)):
                    group = list(grouper)
                    if not title or len(title) < self.min_title_length or len(group) < 2:
                        continue
                    for dropable in map(operator.itemgetter(2), group[0:len(group) - 1]):
                        output.write_tsv(dropable)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class DOAJFiltered(DOAJTask):
    """
    Filter DOAJ by ISSN in assets. Slow.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            'dump': DOAJDump(date=self.date),
            'blacklist': DOAJIdentifierBlacklist(date=self.date),
        }

    @timed
    def run(self):
        identifier_blacklist = load_set_from_target(self.input().get('blacklist'))
        excludes = load_set_from_file(self.assets('028_doaj_filter.tsv'),
                                      func=lambda line: line.replace("-", ""))

        with self.output().open('w') as output:
            with self.input().get('dump').open() as handle:
                for line in handle:
                    record, skip = json.loads(line), False
                    if record['_source']['id'] in identifier_blacklist:
                        continue
                    for issn in record["_source"]["index"]["issn"]:
                        issn = issn.replace("-", "").strip()
                        if issn in excludes:
                            skip = True
                            break
                    if skip:
                        continue
                    output.write(line)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))


class DOAJIntermediateSchema(DOAJTask):
    """
    Convert to intermediate schema via span.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'span-import': Executable(name='span-import', message='http://git.io/vI8NV'),
                'input': DOAJFiltered(date=self.date)}

    @timed
    def run(self):
        output = shellout("span-import -i doaj {input} | pigz -c > {output}", input=self.input().get('input').path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'))


class DOAJExport(DOAJTask):
    """
    Export to various formats
    """
    date = ClosestDateParameter(default=datetime.date.today())
    format = luigi.Parameter(default='solr5vu3', description='export format')

    def requires(self):
        return DOAJIntermediateSchema(date=self.date)

    def run(self):
        output = shellout("span-export -o {format} <(unpigz -c {input}) | pigz -c > {output}", format=self.format, input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        extensions = {
            'solr5vu3': 'ldj.gz',
            'formeta': 'form.gz',
        }
        return luigi.LocalTarget(path=self.path(ext=extensions.get(self.format, 'gz')))


class DOAJISSNList(DOAJTask):
    """
    A list of DOAJ ISSNs.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'input': DOAJIntermediateSchema(date=self.date),
                'jq': Executable(name='jq', message='http://git.io/NYpfTw')}

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        shellout("""jq -r '.["rft.issn"][]?' <(unpigz -c {input}) >> {output} """, input=self.input().get('input').path, output=stopover)
        shellout("""jq -r '.["rft.eissn"][]?' <(unpigz -c {input}) >> {output} """, input=self.input().get('input').path, output=stopover)
        output = shellout("""sort -u {input} > {output} """, input=stopover)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class DOAJDOIList(DOAJTask):
    """
    An best-effort list of DOAJ DOIs.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'input': DOAJIntermediateSchema(date=self.date),
                'jq': Executable(name='jq', message='http://git.io/NYpfTw')}

    @timed
    def run(self):
        output = shellout("""jq -r '.doi' <(unpigz -c {input}) | grep -v "null" | grep -o "10.*" 2> /dev/null | sort -u > {output} """,
                          input=self.input().get('input').path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301,C0103,W0614,W0401,E0202

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Various utilities.
"""

from __future__ import print_function

from future import standard_library
standard_library.install_aliases()
from builtins import zip
from builtins import map
from builtins import range
from builtins import object
import collections
import io as StringIO
import errno
import functools
import hashlib
import itertools
import json
import logging
import operator
import os
import pprint
import random
import string
import sys
import tempfile

import luigi
import requests
from dateutil import relativedelta

import backoff
from siskin import __version__

logger = logging.getLogger('siskin')


class SetEncoder(json.JSONEncoder):
    """ Helper to encode python sets into JSON lists. """

    def default(self, obj):
        """ Decorate call to standard implementation. """
        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)


def date_range(start_date, end_date, increment, period):
    """
    Generate `date` objects between `start_date` and `end_date` in `increment`
    `period` intervals.
    """
    result = []
    nxt = start_date
    delta = relativedelta.relativedelta(**{period: increment})
    while nxt <= end_date:
        result.append(nxt)
        nxt += delta
    return result


def iterfiles(directory='.', fun=None):
    """
    Yield paths below a directory, optionally filter the path through a function
    given in `fun`.
    """
    if fun is None:
        fun = lambda path: True
    for root, _, files in os.walk(directory):
        for f in files:
            path = os.path.join(root, f)
            if fun(path):
                yield path


def random_string(length=16):
    """
    Return a random string(upper and lowercase letters) of length `length`,
    defaults to 16.
    """
    return ''.join(random.choice(string.letters) for _ in range(length))


def pairwise(obj):
    """ Iterator over a iterable in steps of two. """
    iterable = iter(obj)
    return zip(iterable, iterable)


def nwise(iterable, n=2):
    """
    Generalized: func: `pairwise`.
    Split an iterable after every `n` items.
    """
    i = iter(iterable)
    piece = tuple(itertools.islice(i, n))
    while piece:
        yield piece
        piece = tuple(itertools.islice(i, n))


def dictcheck(obj, contains=None, missing=None):
    """
    For a dictionary, check if it contains *values* for the keys given in
    contains and at the same time it does not have values for keys, which are
    given in missing.

    >>> dictcheck({"name": "x"}, contains=["name"])
    True

    >>> dictcheck({"name": "x"}, contains=["name"], missing=["key"])
    True

    >>> dictcheck({"name": "x", "key": "1234"}, contains=["name"], missing=["key"])
    False

    >>> dictcheck({"key": None}, missing=["key"])
    True

    >>> dictcheck({}, missing=["key"])
    True

    """
    if not isinstance(obj, dict):
        raise ValueError('dictionary only')
    if contains is None:
        contains = []
    if missing is None:
        missing = []

    for key in contains:
        if key not in obj:
            return False
        if not bool(operator.itemgetter(key)(obj)):
            return False

    for key in missing:
        if key not in obj:
            continue
        if bool(operator.itemgetter(key)(obj)):
            return False

    return True


def get_task_import_cache():
    """
    Load `taskname: modulename` mappings from dictionary. Return a tuple containing
    the dictionary and the path to the cache file.
    """
    task_import_cache = None
    path = os.path.join(tempfile.gettempdir(),
                        'siskin_task_import_cache_%s' % __version__)
    if not os.path.exists(path):
        logger.debug("creating task import cache at %s", path)
        from siskin.cacheutils import _write_task_import_cache
        _write_task_import_cache(path)

    with open(path) as handle:
        try:
            task_import_cache = json.load(handle)
        except Exception as err:
            print("failed to load task import cache, remove %s, then try again" %
                  path, file=sys.stderr)
            sys.exit(1)

    return task_import_cache, path


def load_set_from_target(target):
    """
    Given a luigi.LocalTarget, load each line of the file into a set.
    """
    s = set()
    with target.open() as handle:
        for line in handle:
            s.add(line.strip())
    return s


def load_set_from_file(filename, func=lambda v: v):
    """
    Given a filename, load each non-empty line into a set.
    """
    s = set()
    with open(filename) as handle:
        for line in map(str.strip, handle):
            if not line:
                continue
            s.add(func(line))
    return s


class URLCache(object):
    """
    A simple URL * content * cache. Stores everything on the filesystem. Content
    is first written to a temporary file and then renamed. With concurrent
    requests for the same URL, the last one wins. Raises exception on any HTTP
    status >= 400. Retries supported.

    >> > cache = URLCache()
    >> > cache.get_cache_file("https://www.google.com")
    /tmp / ef / 7e / fc / ef7efc9839c3ee036f023e9635bc3b056d6ee2d

    >> > cache.is_cached("https://www.google.com")
    False

    >> > page = cache.get("https://www.google.com")
    >> > page[:15]
    '<!doctype html>'

    >> > cache.is_cached("https://www.google.com")
    True
    """

    def __init__(self, directory=None, max_tries=12):
        """
        If `directory` is not explictly given, all files will be stored under
        the temporary directory. Requests can be retried, if they resulted in
        a non 200 HTTP status code. The server might send a HTTP 500 (Internal
        Server Error), even if it really is a HTTP 503 (Service Unavailable).
        We therefore treat HTTP 500 errors as something to retry on,
        `max_tries` times.
        """
        self.directory = directory or tempfile.gettempdir()
        self.sess = requests.session()
        self.max_tries = max_tries

    def get_cache_file(self, url):
        """
        Return the cache file path for a URL. This will create the parent
        directories, if necessary.
        """
        digest = hashlib.sha1(url).hexdigest()
        d0, d1, d2 = digest[:2], digest[2:4], digest[4:6]
        path = os.path.join(self.directory, d0, d1, d2)

        if not os.path.exists(path):
            try:
                os.makedirs(path)
            except OSError as e:
                if e.errno == errno.EEXIST:
                    pass
                else:
                    raise
        return os.path.join(path, digest)

    def is_cached(self, url):
        return os.path.exists(self.get_cache_file(url))

    def get(self, url):
        """
        Return URL, either from cache or the web.
        """

        @backoff.on_exception(backoff.expo, RuntimeError, max_tries=self.max_tries)
        def fetch(url):
            """
            Nested function, so we can configure number of retries.
            """
            r = self.sess.get(url, timeout=600)
            if r.status_code >= 400:
                raise RuntimeError('%s on %s' % (r.status_code, url))
            with tempfile.NamedTemporaryFile(delete=False) as output:
                output.write(r.text.encode('utf-8'))
            os.rename(output.name, self.get_cache_file(url))

        if not self.is_cached(url):
            fetch(url)

        with open(self.get_cache_file(url)) as handle:
            return handle.read()


class ElasticsearchMixin(luigi.Task):
    """ A small mixin for tasks that require an ES connection. """
    es_host = luigi.Parameter(default='localhost', significant=False,
                              description='elasticsearch host')
    es_port = luigi.IntParameter(default=9200, significant=False,
                                 description='elasticsearch port')



# coding: utf-8
# pylint: disable=C0301

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Common tasks.
"""

import datetime
import email.utils as eut
import hashlib
import os
import pipes

import luigi
import requests

from gluish.common import Executable
from gluish.format import TSV
from gluish.utils import shellout
from siskin.task import DefaultTask
from siskin.utils import iterfiles, random_string


class CommonTask(DefaultTask):
    """
    A base class for common classes. These artefacts will be written to the
    systems tempdir.
    """
    TAG = 'common'


class Directory(luigi.Task):
    """ Create directory or fail. """
    path = luigi.Parameter(description='directory to create')

    def run(self):
        try:
            os.makedirs(self.path)
        except OSError as err:
            if err.errno == 17:
                # file exists, this can happen in parallel execution evns
                pass
            else:
                raise RuntimeError(err)

    def output(self):
        return luigi.LocalTarget(self.path)


class FTPMirror(CommonTask):
    """
    A generic FTP directory sync. Required lftp (http://lftp.yar.ru/).
    The output of this task is a single file, that contains the paths
    to all the mirrored files.
    """
    host = luigi.Parameter()
    username = luigi.Parameter(default='anonymous')
    password = luigi.Parameter(default='')
    pattern = luigi.Parameter(default='*', description="e.g. '*leip_*.zip'")
    base = luigi.Parameter(default='.')
    indicator = luigi.Parameter(default=random_string())
    max_retries = luigi.IntParameter(default=5, significant=False)
    timeout = luigi.IntParameter(
        default=10, significant=False, description='timeout in seconds')
    exclude_glob = luigi.Parameter(
        default="", significant=False, description='globs to exclude')

    def requires(self):
        return Executable(name='lftp', message='http://lftp.yar.ru/')

    def run(self):
        """ The indicator is always recreated, while the subdir
        for a given (host, username, base, pattern) is just synced. """
        base = os.path.dirname(self.output().path)
        subdir = hashlib.sha1('{host}:{username}:{base}:{pattern}'.format(
            host=self.host, username=self.username, base=self.base,
            pattern=self.pattern)).hexdigest()

        # target is the root of the mirror
        target = os.path.join(base, subdir)
        if not os.path.exists(target):
            os.makedirs(target)

        exclude_glob = ""
        if not self.exclude_glob == "":
            exclude_glob = "--exclude-glob %s" % self.exclude_glob

        command = """lftp -u {username},{password}
        -e "
            set sftp:auto-confirm yes;
            set net:max-retries {max_retries};
            set net:timeout {timeout};
            set mirror:parallel-directories 1;
            set ssl:verify-certificate no;
            set ftp:ssl-protect-data true;

        mirror --verbose=0 --only-newer {exclude_glob} -I {pattern} {base} {target}; exit" {host}"""

        shellout(command, host=self.host, username=pipes.quote(self.username),
                 password=pipes.quote(self.password),
                 pattern=pipes.quote(self.pattern),
                 target=pipes.quote(target),
                 base=pipes.quote(self.base),
                 max_retries=self.max_retries,
                 timeout=self.timeout,
                 exclude_glob=exclude_glob)

        with self.output().open('w') as output:
            for path in iterfiles(target):
                self.logger.debug("Mirrored: %s", path)
                output.write_tsv(path)

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True), format=TSV)


class FTPFile(CommonTask):
    """ Just require a single file from an FTP server. """
    host = luigi.Parameter()
    username = luigi.Parameter()
    password = luigi.Parameter()
    filepath = luigi.Parameter()

    def requires(self):
        return Executable(name='lftp')

    def run(self):
        command = """lftp -u {username},{password}
        -e "set net:max-retries 5; set net:timeout 10; get -c
        {filepath} -o {output}; exit" {host}"""

        output = shellout(command, host=self.host,
                          username=pipes.quote(self.username),
                          password=pipes.quote(self.password),
                          filepath=pipes.quote(self.filepath))
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True, ext=None))


class HTTPDownload(CommonTask):
    """
    Download a file via HTTP, read out the HTTP Last-modified header and use it as filename.
    """
    url = luigi.Parameter(description='pass this parameter')

    def filename(self):
        """
        Returns the name of the output file. This helper relies on network connectivity.
        """
        r = requests.head(self.url)
        if r.status_code != 200:
            raise RuntimeError('%s on %s' % (r.status_code, self.url))
        value = r.headers.get('Last-Modified')
        if value is None:
            raise RuntimeError('HTTPDownload relies on HTTP Last-Modified header at the moment')
        parsed_date = eut.parsedate(value)
        if parsed_date is None:
            raise RuntimeError('could not parse Last-Modifier header')
        last_modified_date = datetime.date(*parsed_date[:3])
        digest = hashlib.sha1(self.url).hexdigest()
        return '%s-%s.file' % (digest, last_modified_date.isoformat())

    def run(self):
        """
        We try just once. TODO(miku): Some retry bracket.
        Last-Modified date format: Wed, 25 Jan 2017 14:04:59 GMT
        """
        output = shellout(""" curl --fail "{url}" > {output} """, input=self.url)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.filename())
